//
// turing - simulation of Turing machines
// Copyright (C) 2009,2010  Joachim Reichel <joachim.reichel@gmx.de>
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//

/// \file
/// \brief Computation of Goedel numbers.

#ifndef TURING_GOEDEL_H
#define TURING_GOEDEL_H

#include "types.h"
#include "turing_machine.h"

#include <string>
#include <vector>
#include <ostream>
#include <istream>

namespace TM {



/// Computes the Goedel number of a Turing Machine.
///
/// See #goedel(std::ostream&, const Turing_machine&, Encoding, bool) for a
/// variant that writes the Goedel number directly to a stream.
///
/// \param turing_machine   The Turing machine for which the Goedel number is
///                         to be computed.
/// \param encoding         The encoding to be used.
/// \param incomplete       If \c true, Turing machines with initial state
///                         different from 0 or halting state different from 1
///                         are accepted.
/// \return                 The computed goedel number.
std::string goedel(
    const Turing_machine& turing_machine, Encoding encoding,
    bool incomplete = false);

/// Computes the Goedel number of a Turing Machine.
///
/// This method writes the computed Goedel number to the given stream and
/// returns the length of the Goedel number.
///
/// See #goedel(const Turing_machine&, Encoding, bool) for a variant that
/// writes the Goedel number to a string.
///
/// \param stream           The stream to write the Goedel number to.
/// \param turing_machine   The Turing machine for which the Goedel number is
///                         to be computed.
/// \param encoding         The encoding to be used.
/// \param incomplete       If \c true, Turing machines with initial state
///                         different from 0 or halting state different from 1
///                         are accepted.
/// \return                 The length of the computed goedel number.
size_t goedel(
    std::ostream& stream, const Turing_machine& turing_machine,
    Encoding encoding, bool incomplete = false);

/// Constructs a Turing machine from a Goedel number.
///
/// See #ungoedel(std::istream&, Encoding, const Alphabet&) for a variant that
/// reads the Goedel number directly from a stream.
///
/// \param goedel     The Goedel number.
/// \param encoding   The encoding scheme, of the Goedel number.
/// \param alphabet   The alphabet of the Turing machine.
/// \return           The constructed Turing machine, or NULL on failure.
Turing_machine* ungoedel(
    const std::string& goedel, Encoding encoding, const Alphabet& alphabet);

/// Constructs a Turing machine from a Goedel number.
///
/// This method reads the Goedel number from the given stream.
///
/// See #ungoedel(const std::string&, Encoding, const Alphabet&) for a variant
/// that reads the Goedel number from a string.
///
/// \param stream     The stream to read the Goedel number from.
/// \param encoding   The encoding scheme of the Goedel number.
/// \param alphabet   The alphabet of the Turing machine.
/// \return           The constructed Turing machine, or NULL on failure.
Turing_machine* ungoedel(
    std::istream& stream, Encoding encoding, const Alphabet& alphabet);



} // namespace TM

#endif // TURING_GOEDEL_H
