//
// turing - simulation of Turing machines
// Copyright (C) 2009,2010  Joachim Reichel <joachim.reichel@gmx.de>
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//

/// \file
/// \brief The #TM::Tape class.

#ifndef TURING_TAPE_H
#define TURING_TAPE_H

#include "types.h"
#include "alphabet.h"

#include <vector>
#include <ostream>

namespace TM {



/// Storage of a Turing machine
///
/// A tape is a sequence of cells each of which can hold exactly one
/// character. The tape is unbounded on both sides.
///
/// With except of the input, all cells of the tape initially contain the Blank
/// character.
class Tape {

public:
    /// Constructor
    ///
    /// \param initial_content   The initial content of the tape. Must not
    ///                          contain the Blank character. The content
    ///                          starts at index 0 and ends just before
    ///                          index |\p initial_content|.
    Tape( const Character* initial_content = "");


    /// Returns the character of a cell
    inline Character get_character( ptrdiff_t index) const
    {
        if( index >= 0) {
            if( index >= static_cast<ptrdiff_t>( m_right_tape.size()))
                m_right_tape.resize( 2 * index, Blank);
            TM_EXPENSIVE_CHECK(
                index < static_cast<ptrdiff_t>( m_right_tape.size()));
            return m_right_tape[index];
        }

        index = ~index;
        TM_EXPENSIVE_CHECK( index >= 0);
        if( index >= static_cast<ptrdiff_t>( m_left_tape.size()))
            m_left_tape.resize( 2 * index, Blank);
        TM_EXPENSIVE_CHECK(
            index < static_cast<ptrdiff_t>( m_left_tape.size()));
        return m_left_tape[index];
    }

    /// Sets the character of a cell
    ///
    /// \pre   There was an earlier call of get_character() with the same value
    ///        \p index.
    inline void set_character( ptrdiff_t index, Character c)
    {
        if( index >= 0) {
            TM_EXPENSIVE_CHECK(
                index < static_cast<ptrdiff_t>( m_right_tape.size()));
            m_right_tape[index] = c;
        } else {
            TM_EXPENSIVE_CHECK(
                ~index < static_cast<ptrdiff_t>( m_left_tape.size()));
            m_left_tape[~index] = c;
        }

        if( c == Blank) {
            m_boundaries_uptodate = false;
            return;
        }

        if( index < m_left_boundary)
            m_left_boundary = std::min( m_left_boundary, index);
        else if( index > m_right_boundary)
            m_right_boundary = std::max( m_right_boundary, index);
    }

    /// Checks whether the tape content is valid
    ///
    /// The tape content is valid if it contains only characters from the
    /// alphabet.
    ///
    /// \param alphabet   An alphabet to check the tape content against.
    /// \return           \c true if the tape content is valid,
    ///                   otherwise \c false.
    bool is_valid( const Alphabet& alphabet) const;

    /// Returns index of left-most non-Blank character
    ptrdiff_t get_left_boundary()  const;

    /// Returns index of right-most non-Blank character
    ptrdiff_t get_right_boundary() const;

    /// Equality operator
    bool operator==( const Tape& rhs) const;

    /// Inequality operator
    bool operator!=( const Tape& rhs) const;

    /// Output operator
    friend std::ostream& operator<<( std::ostream& out, const Tape& tape);

private:
    /// Updates \c m_left_boundary and \c m_right_boundary.
    ///
    /// \post   \c m_left_boundary is the index of the left-most non-Blank
    ///         character, \c m_right_boundary is the index of the right-most
    ///         non-Blank character.
    void update_boundaries() const;

    /// The cells with non-negative index
    mutable std::vector<Character> m_right_tape;

    /// The cells with negative index
    mutable std::vector<Character> m_left_tape;

    /// Indicates the index of left-most non-Blank character
    ///
    /// This value might be out of date if \c m_boundaries_uptodate is \c false.
    ///
    /// \invariant   There is no non-Blank character left of this index
    mutable ptrdiff_t m_left_boundary;

    /// Indicates the index of right-most non-Blank character
    ///
    /// This value might be out of date if \c m_boundaries_uptodate is \c false.
    ///
    /// \invariant   There is no non-Blank character right of this index
    mutable ptrdiff_t m_right_boundary;

    /// Indicates whether \c m_left_boundary and \c m_right_boundary are
    /// up-to-date.
    mutable bool m_boundaries_uptodate;
};

/// Output operator
std::ostream& operator<<( std::ostream& out, const Tape& tape);



} // namespace TM

#endif // TURING_TAPE_H
