//
// turing - simulation of Turing machines
// Copyright (C) 2009,2010  Joachim Reichel <joachim.reichel@gmx.de>
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//

/// \file
/// \brief The #TM::Transition class.

#ifndef TURING_TRANSITION_H
#define TURING_TRANSITION_H

#include "types.h"

#include <ostream>

namespace TM {



/// %Transition of a Turing machine
///
/// A transition of a Turing machine is a quintuple (q, c, q', c', d), where q
/// and q' are states, c and c' are characters, and d is a direction. If the
/// Turing machine is in state q and its head on a tape cell with character c,
/// the Turing machine switches to state q', writes c' to the tape cell, and
/// moves its head as indicated by d.
class Transition {

public:
    /// Constructor
    Transition( State state, Character character, State new_state,
        Character new_character, Direction direction);

    /// Returns the state q.
    inline State     get_state()         const { return m_state; }

    /// Returns the character c.
    inline Character get_character()     const { return m_character; }

    /// Returns the new state q'.
    inline State     get_new_state()     const { return m_new_state; }

    /// Returns the new character c'.
    inline Character get_new_character() const { return m_new_character; }

    /// Returns the direction d.
    inline Direction get_direction()     const { return m_direction; }

    /// Checks whether the transition matches a given state and character
    ///
    /// \param state       The given state.
    /// \param character   The given character.
    /// \return            \c true if q equals \p state and c equals
    ///                    \p character, otherwise \c false.
    inline bool matches( const State& state, const Character& character) const
    {
        return (state == m_state) && (character == m_character);
    }

    /// Equality operator
    bool operator==( const Transition& rhs) const;

    /// Inequality operator
    bool operator!=( const Transition& rhs) const;

    /// Output operator
    friend std::ostream& operator<<(
        std::ostream& out, const Transition& transition);

private:
    State     m_state;
    State     m_new_state;
    Character m_character;
    Character m_new_character;
    Direction m_direction;
};

/// Output operator
std::ostream& operator<<( std::ostream& out, const Transition& transition);



} // namespace TM

#endif // TURING_TRANSITION_H
