//
// turing - simulation of Turing machines
// Copyright (C) 2009,2010  Joachim Reichel <joachim.reichel@gmx.de>
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//

#include "turing_machine.h"

#include <algorithm>

namespace TM {



Turing_machine::Turing_machine(
    State initial, State halting, const Alphabet& alphabet)
  : m_initial_state( initial),
    m_halting_state( halting),
    m_alphabet( alphabet),
    m_state_to_index_map_uptodate( true)
{
    TM_ALWAYS_CHECK(
        alphabet.get_character_index( Blank) != static_cast<size_t>( -1));
}

void Turing_machine::add_transition( const Transition& transition)
{
    State state = transition.get_state();
    Character character = transition.get_character();
    Character new_character = transition.get_new_character();
    TM_ALWAYS_CHECK( m_alphabet.get_character_index( character)
        != static_cast<size_t>( -1));
    TM_ALWAYS_CHECK( m_alphabet.get_character_index( new_character)
        != static_cast<size_t>( -1));
    TM_ALWAYS_CHECK( find_transition( state, character)
        == static_cast<size_t>( -1));

    m_transition_table.push_back( transition);
    m_state_to_index_map_uptodate = false;
}

size_t Turing_machine::find_transition(
    const State& state, const Character& character) const
{
    size_t index
        = m_state_to_index_map_uptodate && (state < m_state_to_index.size())
          ? m_state_to_index[state] : 0;
    size_t size = m_transition_table.size();

    while( (index < size)
        && ! m_transition_table[index].matches( state, character))
        ++index;

    if( index >= size)
        return static_cast<size_t>( -1);

    return index;
}

size_t Turing_machine::get_nr_of_states() const
{
    return get_states().size();
}

std::set<State> Turing_machine::get_states() const
{
    std::set<State> s;
    size_t size = get_nr_of_transitions();
    for( size_t index = 0; index < size; ++index) {
        const Transition& transition = get_transition( index);
        s.insert( transition.get_state());
        s.insert( transition.get_new_state());
    }
    return s;
}

void Turing_machine::add_transitions_for_fixed_string(
    size_t initial, size_t stride, const Character* s)
{
    size_t state = initial;
    while( *s) {
        add_transition(
            Transition( state, Blank, state+stride, *s, Right));
        ++s;
        state += stride;
    }
}

class Compare_transition
{
public:

    Compare_transition( const Alphabet& alphabet) : m_alphabet( alphabet) {};

    bool operator() (const Transition& lhs, const Transition& rhs)
    {
        return  ( lhs.get_state() < rhs.get_state())
            || (( lhs.get_state() == rhs.get_state())
             && ( lhs.get_character() < rhs.get_character()));
    }

private:
    const Alphabet m_alphabet;
};

void Turing_machine::optimize()
{
    Compare_transition compare_transition( m_alphabet);

    if( m_transition_table.size() == 0) {
        m_state_to_index.resize( 0);
        m_state_to_index_map_uptodate = true;
        return;
    }

    std::sort(
        m_transition_table.begin(), m_transition_table.end(),
        compare_transition);
    State max_state = (*--m_transition_table.end()).get_state();
    m_state_to_index.resize( max_state + 1, static_cast<State>( -1));

    State current_state = static_cast<State>( -1);
    for( size_t index = 0; index < get_nr_of_transitions(); ++index) {
        State state = m_transition_table[index].get_state();
        if( state != current_state) {
            m_state_to_index[state] = index;
            current_state = state;
        }
    }

    m_state_to_index_map_uptodate = true;
}

bool Turing_machine::operator==( const Turing_machine& rhs) const
{

    if (( m_initial_state != rhs.m_initial_state)
     && ( m_halting_state != rhs.m_halting_state)
     && ( m_alphabet != rhs.m_alphabet))
        return false;

    Turing_machine tm_lhs = *this;
    Turing_machine tm_rhs = rhs;
    tm_lhs.optimize();
    tm_rhs.optimize();
    return tm_lhs.m_transition_table == tm_rhs.m_transition_table;
}

bool Turing_machine::operator!=( const Turing_machine& rhs) const
{
    return !( *this == rhs);
}

std::ostream& operator<<(std::ostream& stream, const Turing_machine& tm)
{
    size_t size = tm.get_nr_of_transitions();
    for( size_t index = 0; index < size; ++index)
        stream << tm.get_transition( index) << std::endl;
    return stream;
}



} // namespace TM
